import cv2


def get_circles(
    image,
    min_dist: int,
    param2: int,
    min_radius: int,
    max_radius: int,
):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    # detect circles in the image
    circles = cv2.HoughCircles(
        gray,
        cv2.HOUGH_GRADIENT,
        1.5,
        min_dist,
        None,
        200,
        param2,
        min_radius,
        max_radius,
    )
    return circles



import cv2
import numpy as np


def get_vertices(
    image,
    max_corners: int = 1,
    quality_level: float = 0.01,
    min_distance: int = 100,
):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    corners = cv2.goodFeaturesToTrack(gray, max_corners, quality_level, min_distance)
    return np.intp(corners)



def get_field_of_view(focal_length: float, sensor_width: float, distance: float):
    return sensor_width / focal_length * distance


def get_pixel_per_mm(field_of_view: float, pixel_width: int):
    return pixel_width / field_of_view



import os

# For github actions
CI_MYSQL_CONFIG = dict(
    host="127.0.0.1",
    port=3306,
    user="root",
    password="root",
)


MYSQL_CONFIG = dict(
    host="raspberrypi.local",
    port=3306,
    user="yuchi",
    password="raspberrypi",
)

if os.environ.get("CI"):
    MYSQL_CONFIG = CI_MYSQL_CONFIG



import numpy as np


class Coordinate(np.ndarray):
    def __new__(cls, x, y, z):
        obj = np.array([x, y, z], dtype=float).view(cls)
        return obj

    def __init__(self, x, y, z):
        self.x = self[0]
        self.y = self[1]
        self.z = self[2]

    def __repr__(self):
        return f"Coordinate(x={self.x}, y={self.y}, z={self.z})"

    def __add__(self, other):
        return Coordinate(*super().__add__(other))

    def __sub__(self, other):
        return Coordinate(*super().__sub__(other))

    def __mul__(self, other):
        return Coordinate(*super().__mul__(other))

    def __truediv__(self, other):
        return Coordinate(*super().__truediv__(other))

    def distance_to(self, other) -> float:
        return np.linalg.norm(self - other)

    def unique_key(self) -> str:
        return f"{self.x},{self.y},{self.z}"



from cnceye.coordinate import Coordinate
from cnceye.camera import Camera
from .pixel import get_field_of_view, get_pixel_per_mm
import numpy as np
from cnceye.line import get_lines, Line
from cnceye.vertex import get_vertices
from typing import List
import mysql.connector
from mysql.connector.errors import IntegrityError
from cnceye.config import MYSQL_CONFIG


class SingleImage:
    def __init__(self, image, center_coordinate: Coordinate, camera: Camera) -> None:
        self.image = image
        # z coordinate is the height of the object
        self.center = center_coordinate
        self.camera = camera

    def get_opencv_origin(self, image, distance: float) -> Coordinate:
        # opencv origin is at top left corner
        pixel_per_mm = self.pixel_per_mm(distance)
        diff_in_pixel = (-image.shape[1] / 2, image.shape[0] / 2, 0)
        diff_in_mm = np.array([x / pixel_per_mm for x in diff_in_pixel])
        return self.center + diff_in_mm

    def pixel_per_mm(self, distance: float) -> float:
        field_of_view = get_field_of_view(
            self.camera.focal_length, self.camera.sensor_width, distance
        )
        return get_pixel_per_mm(field_of_view, self.image.shape[1])

    def from_opencv_coord(self, distance: float, opencv_xy: tuple) -> Coordinate:
        pixel_per_mm = self.pixel_per_mm(distance)
        opencv_origin = self.get_opencv_origin(self.image, distance)

        return Coordinate(
            opencv_origin.x + opencv_xy[0] / pixel_per_mm,
            opencv_origin.y - opencv_xy[1] / pixel_per_mm,
            self.center.z,
        )

    def from_pixel_length(self, distance: float, pixel_length) -> float:
        pixel_per_mm = self.pixel_per_mm(distance)
        return pixel_length / pixel_per_mm

    def lines(self, distance: float) -> List[Line] or None:
        line_array = get_lines(self.image, max_line_gap=50)
        if line_array is None:
            return None

        lines = []
        for line in line_array:
            x1, y1, x2, y2 = line[0]
            start = self.from_opencv_coord(distance, (x1, y1))
            end = self.from_opencv_coord(distance, (x2, y2))
            lines.append(Line(start, end))

        return lines

    def vertex(self, distance: float) -> Coordinate:
        vertices = get_vertices(self.image)
        if vertices is None:
            return None

        x, y = vertices[0][0]
        return self.from_opencv_coord(distance, (x, y))

    def add_real_coordinate(self, distance):
        real_coordinate = self.vertex(distance)
        point_id = self.center.unique_key()

        cnx = mysql.connector.connect(**MYSQL_CONFIG, database="coord")
        cursor = cnx.cursor()

        update_query = """
            UPDATE point
            SET rx = %s, ry = %s, rz = %s
            WHERE point_id = %s
        """
        try:
            data = (real_coordinate.x, real_coordinate.y, real_coordinate.z, point_id)
            cursor.execute(update_query, data)
        except IntegrityError:
            print("Error: unable to update points")

        cnx.commit()
        cursor.close()
        cnx.close()



import cv2
import json
import numpy as np


def undistort_img(img, camera_data_path: str):
    h, w = img.shape[:2]
    camera_matrix, dist_coefs = get_camera_data(camera_data_path)
    newcameramtx, roi = cv2.getOptimalNewCameraMatrix(
        camera_matrix, dist_coefs, (w, h), 1, (w, h)
    )

    dst = cv2.undistort(img, camera_matrix, dist_coefs, None, newcameramtx)

    # crop and save the image
    x, y, w, h = roi
    dst = dst[y : y + h, x : x + w]
    return dst


def get_camera_data(camera_data_path: str):
    with open(camera_data_path, "r") as f:
        data = json.load(f)
        camera_matrix = np.array(data["camera_matrix"])
        dist_coefs = np.array(data["distortion_coefficients"])

    return camera_matrix, dist_coefs



from cnceye.coordinate import Coordinate
from cnceye.camera import Camera
from cnceye.line import Line
from cnceye.cmm.single import SingleImage
import cv2
import numpy as np
import mysql.connector
from cnceye.config import MYSQL_CONFIG


class AllImages:
    def __init__(self, camera: Camera) -> None:
        self.camera = camera
        self.previous_lines = []
        self.lines = []

    def add_image(self, image, distance: float, center: Coordinate) -> None:
        single = SingleImage(image, center, self.camera)
        lines = single.lines(distance)
        if lines is None:
            return None

        if len(self.previous_lines) == 0:
            self.previous_lines = lines
            return None

        for line in lines:
            is_new_line = True
            for i, previous_line in enumerate(self.previous_lines):
                new_line = line.connect_lines(previous_line)
                if new_line is not None:
                    self.previous_lines[i] = new_line
                    is_new_line = False
                    break

            if is_new_line:
                self.previous_lines.append(line)

    def save_image(self, path: str) -> None:
        entire_image = np.asarray([[[0, 0, 0]] * 1200] * 1000, dtype=np.uint8)
        for line in self.lines:
            start = line.start
            end = line.end
            cv2.line(
                entire_image,
                (int((start.x + 100) * 5), int((-start.y + 100) * 5)),
                (int((end.x + 100) * 5), int((-end.y + 100) * 5)),
                (255, 255, 255),
                1,
            )
            cv2.putText(
                entire_image,
                f"{line.get_length():.3f}",
                (
                    int((start.x + end.x + 200) * 5 / 2),
                    int((-start.y - end.y + 200) * 5 / 2),
                ),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.4,
                (105, 145, 209),
                1,
            )
        cv2.imwrite(path, entire_image)

    def fetch_real_coordinates(self) -> dict:
        cnx = mysql.connector.connect(**MYSQL_CONFIG, database="coord")
        cursor = cnx.cursor()

        real_coordinates = {}
        query = """
            SELECT point_id, x, y, z
            FROM point
        """
        cursor.execute(query)
        for r in cursor:
            real_coordinates[r[0]] = Coordinate(r[1], r[2], r[3])

        cursor.close()
        cnx.close()

        return real_coordinates

    def fetch_lines(self):
        cnx = mysql.connector.connect(**MYSQL_CONFIG, database="coord")
        cursor = cnx.cursor()

        lines = []
        query = """
            SELECT a, b
            FROM line
        """
        cursor.execute(query)
        for line in cursor:
            lines.append((line[0], line[1]))

        cursor.close()
        cnx.close()

        return lines

    def add_lines(self):
        real_coordinates = self.fetch_real_coordinates()
        lines = self.fetch_lines()
        for line in lines:
            start = real_coordinates[line[0]]
            end = real_coordinates[line[1]]
            self.lines.append(Line(start, end))



import cv2
import numpy as np
from cnceye.coordinate import Coordinate


class Line:
    def __init__(self, start: Coordinate, end: Coordinate) -> None:
        # start.x < end.x
        if start.x > end.x:
            self.start = end
            self.end = start
        else:
            self.start = start
            self.end = end

    def get_slope(self) -> float:
        if abs(self.start.x - self.end.x) < 0.1:
            return np.inf
        return (self.end.y - self.start.y) / (self.end.x - self.start.x)

    def get_slope_diff(self, other: "Line") -> float:
        if self.get_slope() == other.get_slope():
            return 0
        return abs(self.get_slope() - other.get_slope())

    def get_intercept(self) -> float:
        if self.get_slope() == np.inf:
            return np.inf
        return self.start.y - self.get_slope() * self.start.x

    def get_intercept_diff(self, other: "Line") -> float:
        if self.get_intercept() == other.get_intercept():
            return 0
        return abs(self.get_intercept() - other.get_intercept())

    def get_x(self, y: float) -> float:
        return (y - self.get_intercept()) / self.get_slope()

    def get_y(self, x: float) -> float:
        return self.get_slope() * x + self.get_intercept()

    def get_length(self) -> float:
        return np.sqrt(
            (self.end.x - self.start.x) ** 2 + (self.end.y - self.start.y) ** 2
        )

    def get_intersection(self, other) -> tuple:
        x = (other.get_intercept() - self.get_intercept()) / (
            self.get_slope() - other.get_slope()
        )
        y = self.get_slope() * x + self.get_intercept()
        return (x, y)

    def is_same_straight_line(self, other: "Line") -> bool:
        slope_diff = self.get_slope_diff(other)
        intercept_diff = self.get_intercept_diff(other)
        if self.get_slope() == np.inf and other.get_slope() == np.inf:
            return abs(self.start.x - other.start.x) < 0.2
        return slope_diff < 0.1 and intercept_diff < 0.1

    def is_x_overlapping(self, other: "Line") -> bool:
        return (
            self.start.x <= other.start.x <= self.end.x
            or self.start.x <= other.end.x <= self.end.x
            or other.start.x <= self.start.x <= other.end.x
            or other.start.x <= self.end.x <= other.end.x
        )

    def is_y_overlapping(self, other: "Line") -> bool:
        return (
            self.start.y <= other.start.y <= self.end.y
            or self.end.y <= other.start.y <= self.start.y
            or self.start.y <= other.end.y <= self.end.y
            or self.end.y <= other.end.y <= self.start.y
            or other.start.y <= self.start.y <= other.end.y
            or other.end.y <= self.start.y <= other.start.y
            or other.start.y <= self.end.y <= other.end.y
            or other.end.y <= self.start.y <= other.start.y
        )

    def is_overlapping(self, other: "Line") -> bool:
        if self.get_slope() == np.inf and other.get_slope() == np.inf:
            return self.is_y_overlapping(other)
        if self.get_slope() == 0 and other.get_slope() == 0:
            return self.is_x_overlapping(other)
        return self.is_x_overlapping(other) and self.is_y_overlapping(other)

    def connect_lines(self, other: "Line") -> "Line" or None:
        if self.is_same_straight_line(other) and self.is_overlapping(other):
            return Line(
                Coordinate(
                    min(self.start.x, other.start.x),
                    min(self.start.y, other.start.y),
                    self.start.z,
                ),
                Coordinate(
                    max(self.end.x, other.end.x),
                    max(self.end.y, other.end.y),
                    self.start.z,
                ),
            )

    def __repr__(self) -> str:
        return f"Line({self.start}, {self.end})"


def get_lines(
    image,
    gaussian_blur_size=5,
    canny_low_threshold=100,
    canny_high_threshold=200,
    rho=0.1,
    hough_threshold=100,
    min_line_length=50,
    max_line_gap=200,
):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    blur_gray = cv2.GaussianBlur(gray, (gaussian_blur_size, gaussian_blur_size), 0)
    edges = cv2.Canny(
        blur_gray,
        canny_low_threshold,
        canny_high_threshold,
        apertureSize=3,
        L2gradient=True,
    )
    edges = cv2.dilate(edges, None, iterations=1)
    edges = cv2.erode(edges, None, iterations=1)

    lines = cv2.HoughLinesP(
        edges,
        rho,
        np.pi / 180 * rho,
        hough_threshold,
        minLineLength=min_line_length,
        maxLineGap=max_line_gap,
    )
    return lines
